from plynx.demo.hello_world import GROUP as hello_group
from plynx.demo.types import GROUP as types_group
from plynx.demo.basic_functions import GROUP as basic_group


COLLECTION = [
    hello_group,
    types_group,
    basic_group,
]
